From iris.base_logic.lib Require Import invariants.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import excl.
From lrust.lang Require Import notation.
From gpfsl.logic Require Import view_invariants.
From gpfsl.gps Require Import surface protocols escrows.

Set Default Proof Using "Type".

Definition spawn : val :=
  λ: ["f"],
    let: "c" := Alloc #2 in
    "c" <- #0;; (* Initialize "sent" field to 0 (non-atomically). *)
    Fork ("f" ["c"]);;
    "c".
Definition finish : val :=
  λ: ["c"; "v"],
    "c" +ₗ #1 <- "v";; (* Store data (non-atomically). *)
    "c" <-ʳᵉˡ #1;; (* Signal that we finished (atomically!) *)
    #☠.
Definition join : val :=
  rec: "join" ["c"] :=
    if: !ᵃᶜ"c" then
      (* The thread finished, we can non-atomically load the data. *)
      let: "v" := !("c" +ₗ #1) in
      Free #2 "c";;
      "v"
    else
      "join" ["c"].

(** The CMRA & functor we need. *)
Class spawnG Σ := SpawnG {
  spawn_tokG :> inG Σ (exclR unitO);
  spawn_gpsG :> gpsG Σ unitProtocol;
  spawn_viewG :> view_invG Σ }.

Definition spawnΣ : gFunctors := #[GFunctor (exclR unitO); gpsΣ unitProtocol; view_invΣ].

Instance subG_spawnΣ {Σ} : subG spawnΣ Σ → spawnG Σ.
Proof. solve_inG. Qed.

(** Now we come to the Iris part of the proof. *)
Section proof.
Context `{!noprolG Σ, !spawnG Σ, !view_invG Σ} (N : namespace).
Local Notation vProp := (vProp Σ).
Local Notation tok γ := (own γ (Excl ())).
Implicit Types (c: loc) (Ψ : val → vProp) (v: val).

Definition finishEscrow γc γe γi c Ψ :=
  ([es ⎡ tok γe ⎤ ⇝ ((∃ t, GPS_SWWriter c t () #1 γc) ∗ view_tok γi (1/2)%Qp ∗
                            ∃ v, (c >> 1) ↦ v ∗ Ψ v)])%I.

(* GPS protocol interpretation *)
Definition spawn_interp γe γi Ψ : interpO Σ unitProtocol :=
  (λ _ c γc _ _ v, ∃ b : bool, ⌜v = #b⌝ ∗
       if b then finishEscrow γc γe γi c Ψ else True)%I.

Let IW : loc → gname → time → unitProtocol → val → vProp :=
  (λ _ _ _ _ v, ⌜v = #false⌝)%I.

Definition finish_handle γc γe c Ψ : vProp :=
  (∃ γi t v, (c >> 1) ↦ v ∗
    GPS_vSP_Writer N (spawn_interp γe γi Ψ) IW c t () #0 γc γi ∗
    view_tok γi (1/2)%Qp)%I.

Definition join_handle γc γe c Ψ : vProp :=
  (∃ γi (Ψ': val → vProp) t,
    GPS_vSP_Reader N (spawn_interp γe γi Ψ') IW c t () #0 γc γi
    ∗ ⎡ † c … 2 ∗ tok γe ⎤
    ∗ view_tok γi (1/2)%Qp
    ∗ □ (∀ v, Ψ' v -∗ Ψ v))%I.

Global Instance finish_handle_contractive n γc γe c :
  Proper (pointwise_relation val (dist_later n) ==> dist n) (finish_handle γc γe c).
Proof.
  move => ???. do 3 (apply bi.exist_ne => ?).
  apply bi.sep_ne; [done|]. apply bi.sep_ne; [|done].
  apply GPS_vSP_Writer_contractive; [|done]. destruct n; [done|].
  simpl => ??????. apply bi.exist_ne => b'.
  apply bi.sep_ne; [done|]. destruct b'; [|done]. solve_proper.
Qed.
Global Instance join_handle_ne n γc γe c :
  Proper (pointwise_relation val (dist n) ==> dist n) (join_handle γc γe c).
Proof. solve_proper. Qed.

(** The main proofs. *)
Lemma spawn_spec Ψ e (f : val) tid :
  IntoVal e f →
  {{{ ∀ γc γe c tid', finish_handle γc γe c Ψ -∗ WP f [ #c] in tid' {{ _, True }} }}}
    spawn [e] in tid
  {{{ γc γe c, RET #c; join_handle γc γe c Ψ }}}.
Proof.
  iIntros (<- Φ) "Hf HΦ". rewrite /spawn /=.
  wp_let. wp_alloc l as "Hl" "H†". wp_let.
  iMod (own_alloc (Excl ())) as (γe) "Hγe"; first done.
  rewrite own_loc_na_vec_cons own_loc_na_vec_singleton.
  iDestruct "Hl" as "[Hc Hd]". wp_write.
  iMod (GPS_vSP_Init N (λ γi, spawn_interp γe γi Ψ) (λ _, IW) _ () with "Hc []")
    as (γi γc t) "[[Htok1 Htok2] SW]". { iIntros (???). by iExists false. }
  iDestruct (GPS_vSP_SWWriter_Reader with "SW") as "#R".
  wp_apply (wp_fork with "[SW Htok1 Hf Hd]"); [done|..].
  - iNext. iIntros (tid'). iApply "Hf". iExists _, _, _. by iFrame.
  - iIntros "_". wp_seq. iApply ("HΦ" $! γc γe). iExists _, _, _.
    iFrame "R ∗". auto.
Qed.

Lemma finish_spec Ψ c v γc γe tid
  (DISJ1: (↑histN) ## (↑N : coPset)) (DISJ2: (↑escrowN) ## (↑N : coPset)) :
  {{{ finish_handle γc γe c Ψ ∗ Ψ v }}}
    finish [ #c; v] in tid
  {{{ RET #☠; True }}}.
Proof.
  iIntros (Φ) "[Hfin HΨ] HΦ".
  iDestruct "Hfin" as (γi t v0) "(Hd & SW & Hf)".
  wp_lam. wp_op. wp_write. wp_bind (_ <-ʳᵉˡ _)%E.
  iApply (GPS_vSP_SWWrite_rel N (spawn_interp γe γi Ψ) IW _ (1 / 2)%Qp
            (λ _, True)%I True%I (spawn_interp γe γi Ψ true c γc t tt #0)
            t () () _ #1 with "[$Hf $SW HΨ Hd]"); [done..| |].
  { iSplitR "".
    - iIntros "!>" (t' Ext') "SW _ tok !>". iSplitL ""; [by iIntros "!> _"|].
      iSplitR ""; [|done]. iExists true.
      iSplitL ""; [done|]. iApply escrow_alloc; [solve_ndisj|]. iFrame.
      iSplitL "SW"; iExists _; by iFrame.
    - by iIntros "!> !> $". }
  iIntros "!>" (t') "_". wp_seq. by iApply "HΦ".
Qed.

Lemma join_spec Ψ c γc γe tid
  (DISJ1: (↑histN) ## (↑N : coPset)) (DISJ2: (↑escrowN) ## (↑N : coPset)) :
  {{{ join_handle γc γe c Ψ }}} join [ #c] in tid {{{ v, RET v; Ψ v }}}.
Proof.
  iIntros (Φ) "H HΦ".
  iDestruct "H" as (γi Ψ' t) "(R & (H† & He) & Hj & #HΨ')".
  iLöb as "IH" forall (t). wp_rec. wp_bind (!ᵃᶜ _)%E.
  set P : vProp := (⎡ tok γe ⎤)%I.

  iApply (GPS_vSP_SWRead_acq_dealloc N (spawn_interp γe γi Ψ') IW _ P
            (λ _ _, ▷ ∃ v, (c >> 1) ↦{1} v ∗ Ψ' v)%I
            (spawn_interp γe γi Ψ' false c γc) (1/2)%Qp t () _ #true
            with "[$Hj $R $He]"); [done..| |].
  { iSplitL"".
    - iIntros "!>" (t' [] v' [_ Ext']) "!>". iSplit; last iSplit.
      + iDestruct 1 as (b) "[#Eq ES]".
        destruct b; simpl; [iDestruct "ES" as "#ES"|];
          iModIntro; iSplitL ""; iExists _; by iFrame "Eq".
      + iDestruct 1 as (b) "[#Eq ES]".
        destruct b; simpl; [iDestruct "ES" as "#ES"|];
          iModIntro; iSplitL ""; iExists _; by iFrame "Eq".
      + iDestruct 1 as %?. subst v'. iIntros "!>"; iSplit; [done|by iExists false].
    - iIntros "!>" (t' []) "Ext He $".
      iDestruct 1 as (b) "[% Es]". destruct b; last done.
      iMod (escrow_elim with "[] Es [$He]") as "(SW & >$ & $)"; [solve_ndisj| |done].
      iIntros "[He1 He2]". iCombine "He1" "He2" as "He".
      iDestruct (own_valid with "He") as "%". auto. }
  iIntros "!>" (t' [] v') "(Ext & R & CASE)".
  case (decide (v' = #true)) => ?.
  - subst v'. iDestruct "CASE" as "[HInv Ho]". iDestruct "Ho" as (v) "[Hd HΨ]".
    wp_if. wp_op. wp_read. wp_let. iApply (wp_hist_inv); [done|]. iIntros "Hist".
    iMod (GPS_INV_dealloc _ _ _ true with "Hist HInv") as (t1 s1 v1) "[Hc _]"; [done|].
    iAssert (c ↦∗ [ v1 ; v])%I with "[Hc Hd]" as "Hc".
    { rewrite own_loc_na_vec_cons own_loc_na_vec_singleton. iFrame. }
    wp_free; first done. iApply "HΦ". iApply "HΨ'". done.
  - iDestruct "CASE" as "(Hj & res & He)".
    iDestruct "res" as (b) "[% _]". subst v'. destruct b; [done|].
    wp_if. iApply ("IH" with "R H† He Hj"). auto.
Qed.

Lemma join_handle_impl Ψ1 Ψ2 γc γe c :
  □ (∀ v, Ψ1 v -∗ Ψ2 v) -∗ join_handle γc γe c Ψ1 -∗ join_handle γc γe c Ψ2.
Proof.
  iIntros "#HΨ Hdl". iDestruct "Hdl" as (γi Ψ' t) "(? & (? & ?) & ? & #HΨ')".
  iExists γi, Ψ', t. iFrame "#∗". iIntros "!# * ?".
  iApply "HΨ". iApply "HΨ'". done.
Qed.
End proof.

Typeclasses Opaque finish_handle join_handle.
