From Coq.QArith Require Import Qcanon.
From iris.bi Require Import fractional.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import auth excl csum frac agree.
From gpfsl.gps Require Import middleware protocols.
From lrust.lifetime Require Import at_borrow.
From lrust.typing Require Import typing.
From lrust.logic Require Import gps.
Set Default Proof Using "Type".

Definition rwlock_st := option (() + lft * frac * positive).

Definition rwlock_stR :=
  optionUR (csumR (exclR unitO) (prodR (prodR (agreeR lftO) fracR) positiveR)).
Definition rwlockR := authR rwlock_stR.
Class rwlockG Σ := RwLockG {
  rwlock_stateG :> inG Σ rwlockR ;
  rwlock_gpsG :> gpsG Σ unitProtocol;
}.

Definition rwlockΣ : gFunctors := #[GFunctor rwlockR; gpsΣ unitProtocol].
Instance subG_rwlockΣ {Σ} : subG rwlockΣ Σ → rwlockG Σ.
Proof. solve_inG. Qed.

Definition rwlock_to_st (st : rwlock_st) : rwlock_stR :=
  match st with
  | None => None
  | Some (inr (ν, q, n)) => Some (Cinr (to_agree ν, q, n))
  | Some (inl _) => Some (Cinl (Excl ()))
  end.

Definition Z_of_rwlock_st (st : rwlock_st) : Z :=
  match st with
  | None => 0
  | Some (inr (_, _, n)) => Zpos n
  | Some _ => -1
  end.

Definition rwlockN := lrustN .@ "rwlock".

Notation rwown γs st := (⎡ own γs (st : rwlockR) ⎤ : vProp _)%I.
Notation st_full st  := (● rwlock_to_st st ⋅ ◯ rwlock_to_st st).
Notation st_global st:= (● rwlock_to_st st).
Notation st_local st := (◯ rwlock_to_st st).

Section rwlock_inv.
  Context `{!typeG Σ, !rwlockG Σ}.
  Local Notation vProp := (vProp Σ).

  (* Cmra operations *)
  Definition reading_st (q : frac) (ν : lft) : rwlockR
    := st_local (Some $ inr (ν, q, xH)).
  Definition writing_st : rwlockR := st_local (Some $ inl ()).

  Lemma rwown_update_write_read ν γs :
    let st' : rwlock_st := Some (inr (ν, (1/2)%Qp, 1%positive)) in
    rwown γs (st_global None)
    ==∗ rwown γs (st_global st') ∗ rwown γs (reading_st (1/2)%Qp ν).
  Proof.
    iIntros (st') "g". rewrite -embed_sep -own_op.
    iMod (own_update with "g") as "$"; [|done]. apply auth_update_alloc.
    rewrite /= -(right_id None op (Some _)). by apply op_local_update_discrete.
  Qed.

  Lemma rwown_update_read ν (n: positive) (q q': frac) γs
    (Hqq' : (q + q')%Qp = 1%Qp) :
    let st : rwlock_st := (Some (inr (ν, q, n))) in
    let st': rwlock_st := Some (inr (ν, (q + (q'/2)%Qp)%Qp, (n + 1)%positive)) in
    rwown γs (st_global st)
    ==∗ rwown γs (st_global st') ∗ rwown γs (reading_st (q'/2)%Qp ν).
  Proof.
    iIntros (st') "g". rewrite -embed_sep -own_op.
    iMod (own_update with "g") as "$"; [|done]. apply auth_update_alloc.
    rewrite Pos.add_comm Qp_plus_comm -pos_op_plus /=
            -{2}(agree_idemp (to_agree _)) -frac_op'
            2!pair_op Cinr_op Some_op.
    rewrite -{2}(right_id None op (Some (Cinr (_, (q' /2)%Qp, _)))).
    apply op_local_update_discrete =>-[Hagv _]. split; [split|done].
    - by rewrite /= agree_idemp.
    - apply frac_valid'. rewrite -Hqq' comm -{2}(Qp_div_2 q').
      apply Qcplus_le_mono_l. rewrite -{1}(Qcanon.Qcplus_0_l (q'/2)%Qp).
      apply Qcplus_le_mono_r, Qp_ge_0.
  Qed.

  Lemma rwown_update_write γs :
    let st' : rwlock_st := Some (inl ()) in
    rwown γs (st_global None) ==∗ rwown γs (st_global st') ∗ rwown γs writing_st.
  Proof.
    iIntros (st') "g". rewrite -embed_sep -own_op.
    iMod (own_update with "g") as "$"; [|done]. apply auth_update_alloc.
    rewrite /= -(right_id None op (Some _)). by apply op_local_update_discrete.
  Qed.

  Lemma rwown_update_write_dealloc γs :
    let st : rwlock_st := Some (inl ()) in
    rwown γs (st_global st) -∗ rwown γs writing_st ==∗ rwown γs (st_global None).
  Proof.
    iIntros (st) "g w". iCombine "g" "w" as "gw".
    iMod (own_update with "gw") as "$"; [|done]. apply auth_update_dealloc.
    rewrite /= -(right_id None op (Some _)). by apply cancel_local_update_unit, _.
  Qed.

  Lemma rwown_global_reading_st st q ν γs:
    rwown γs (st_global st) -∗ rwown γs (reading_st q ν) -∗
    ⌜∃ q' n, st = Some (inr (ν,q',n))
      ∧ ( q' = q ∧ n = 1%positive
        ∨ (∃ q0, q' = (q + q0)%Qp) ∧ n = (1 + Pos.pred n)%positive)⌝.
  Proof.
    iIntros "m r". iCombine "m" "r" as "mr".
    iDestruct (own_valid with "mr") as %[INCL VAL]%auth_both_valid.
    iPureIntro. destruct st as [[[]|[[]]]|].
    - move : INCL =>
        /Some_included [Eq|/csum_included [//|[[?[?[??//]]]|[?[?[?[??//]]]]]]].
      inversion Eq.
    - move : INCL =>
        /Some_included [Eq|/csum_included [//|[[?[?[??//]]]|[a[b[Eqa[Eqb INCL]]]]]]].
      + inversion Eq as [| ?? Eq1 |]. inversion Eq1 as [Eq2].
        inversion Eq2 as [Eq3 ]. simpl in Eq3. apply to_agree_inj in Eq3.
        inversion Eq3. simpl in *. subst. do 2 eexists.
        split; [done|by left].
      + inversion Eqa. inversion Eqb. subst a b.
        apply prod_included in INCL as
          [[INCL%to_agree_included LE1]%prod_included LE2%pos_included].
        inversion INCL. do 2 eexists. split; [done|]. right. split.
        * rewrite /= in LE1. destruct LE1 as [q1 LE1]. exists q1. by rewrite -frac_op'.
        * rewrite -Pplus_one_succ_l Pos.succ_pred => //?. by subst.
    - by apply option_included in INCL as [?|[?[?[?[??]]]]].
  Qed.

  Lemma rwown_update_read_dealloc_1 γs ν q:
    rwown γs (st_global (Some (inr (ν,q,1%positive)))) -∗ rwown γs (reading_st q ν)
    ==∗ rwown γs (st_global None).
  Proof.
    iIntros "m r". iCombine "m" "r" as "mr".
    iMod (own_update with "mr") as "$"; [|done].
    apply auth_update_dealloc.
    rewrite /= -(right_id None op (Some _)). apply cancel_local_update_unit, _.
  Qed.

  Lemma rwown_update_read_dealloc γs ν q q' n:
    rwown γs (st_global (Some (inr (ν,(q + q')%Qp,(1 + n)%positive))))
    -∗ rwown γs (reading_st q' ν)
    ==∗ rwown γs (st_global (Some (inr (ν,q,n)))).
  Proof.
    iIntros "m r". iCombine "m" "r" as "mr".
    iMod (own_update with "mr") as "$"; [|done].
    apply auth_update_dealloc.
    rewrite -frac_op' -pos_op_plus /= (cmra_comm_L q)
            -{1}(agree_idemp (to_agree _)) 2!pair_op Cinr_op Some_op.
    by apply (cancel_local_update_unit (Some _)), _.
  Qed.

  Lemma rwown_global_writing_st st γs:
    rwown γs (st_global st) -∗  rwown γs writing_st -∗
    ⌜st = (Some (inl tt))⌝.
  Proof.
    iIntros "m w". iCombine "m" "w" as "mw".
    iDestruct (own_valid with "mw") as %[INCL VAL]%auth_both_valid.
    iPureIntro. destruct st as [[[]|[[]]]|]; [done|..].
    - move : INCL =>
        /Some_included [Eq|/csum_included [//|[[?[?[?[Eq _]]]]|[?[?[?[??//]]]]]]];
      inversion Eq.
    - move :INCL => /= /option_included [//|[?[? [? [? //]]]]].
  Qed.

  (* GPS protocol definitions *)
  Definition rwlock_interp
    (γs: gname) (α : lft) (tyO: vProp) (tyS: lft → vProp)
    : interpO Σ unitProtocol :=
    (λ pb l γ t _ vs, ∃ st, ⌜vs = #(Z_of_rwlock_st st)⌝ ∗
       (□ (∀ κ q E, ⌜lftE ⊆ E⌝ -∗ &{κ}tyO -∗ q.[κ] ={E}=∗ (□ tyS κ) ∗ q.[κ])) ∗
       if pb
       then rwown γs (st_global st) ∗
            match st return _ with
            | None => (* Not locked. *)
                  &{α}tyO ∗
                  GPS_SWSharedWriter l t () #0 γ ∗
                  GPS_SWSharedReader l t () #0 1%Qp γ
            | Some (inr (ν, q, n)) => (* Locked for read. *)
                  GPS_SWSharedWriter l t () #(Z.pos n) γ ∗
                  ∃ q', □ (1.[ν] ={↑lftN ∪ ↑histN,↑histN}▷=∗ [†ν]) ∗
                    ([†ν] ={↑lftN ∪ ↑histN}=∗ &{α}tyO) ∗
                    (□ tyS (α ⊓ ν)) ∗ ⌜(q + q')%Qp = 1%Qp⌝ ∗
                    q'.[ν] ∗
                    GPS_SWSharedReader l t () #(Z.pos n) q' γ
            | _ => (* Locked for write. *) True
            end
       else True)%I.

  Definition rwlock_noCAS : interpCasO Σ unitProtocol :=
    (λ _ _ _ _ v, ⌜v = #(-1)⌝)%I.

  Definition rwlock_proto' l γ γs κ (tyO: vProp) (tyS: lft → vProp): vProp :=
    (∃ t v, GPS_aSP_Reader rwlockN (rwlock_interp γs κ tyO tyS)
                           rwlock_noCAS l κ t () v γ)%I.
  Definition rwlock_proto l γ γs κ tyO tyS tid ty: vProp :=
    ((▷ □ (tyO ↔ (l +ₗ 1) ↦∗: ty.(ty_own) tid)) ∗
     (▷ □ (∀ α, tyS α ↔ ty.(ty_shr) α tid (l +ₗ 1))) ∗
      rwlock_proto' l γ γs κ tyO tyS)%I.

  Lemma rwlock_interp_comparable γs α tyO tyS l γ t s v (n: Z):
    rwlock_interp γs α tyO tyS true l γ t s v ∨
    rwlock_interp γs α tyO tyS false l γ t s v
    -∗ ⌜∃ vl : lit, v = #vl ∧ lit_comparable n vl⌝.
  Proof.
    iDestruct 1 as "[F|F]"; iDestruct "F" as (st) "[% _]"; subst v;
    destruct st as [[|[[??]?]]|]; simpl;
      iExists _; iPureIntro ;(split; [done|by constructor]).
  Qed.

  Global Instance rwlock_proto_persistent:
    Persistent (rwlock_proto l γ γs κ tyO tyS tid ty) := _.

  Global Instance rwlock_proto_type_ne n l γ γs κ tyO tyS tid :
    Proper (type_dist2 n ==> dist n) (rwlock_proto l γ γs κ tyO tyS tid).
  Proof.
    move => ???.
    apply bi.sep_ne; [|apply bi.sep_ne]; [..|done];
      apply bi.later_contractive; (destruct n; [done|]);
      apply bi.intuitionistically_ne.
    - apply bi.iff_ne; [done|]. apply bi.exist_ne => ?.
      apply bi.sep_ne; [done|by apply ty_own_type_dist].
    - apply bi.forall_ne => ?.
      apply bi.iff_ne; [done|apply ty_shr_type_dist]; [by apply type_dist2_S|done..].
  Qed.

  Global Instance rwlock_proto_ne l γ γs κ tyO tyS tid :
    NonExpansive (rwlock_proto l γ γs κ tyO tyS tid).
  Proof.
    intros n ???. by eapply rwlock_proto_type_ne, type_dist_dist2.
  Qed.

  Lemma rwlock_proto_proper E L ty1 ty2 q :
    eqtype E L ty1 ty2 →
    llctx_interp L q -∗ ∀ l γ γs κ tyO tyS tid, □ (elctx_interp E -∗
    rwlock_proto l γ γs κ tyO tyS tid ty1 -∗ rwlock_proto l γ γs κ tyO tyS tid ty2).
  Proof.
    (* TODO : this proof is essentially [solve_proper], but within the logic.
              It would easily gain from some automation. *)
    rewrite eqtype_unfold. iIntros (Hty) "HL".
    iDestruct (Hty with "HL") as "#Hty". iIntros "* !# #HE H".
    iDestruct ("Hty" with "HE") as "(% & #Hown & #Hshr)".
    iDestruct "H" as "(#EqO & #EqS & $)". iSplit; iIntros "!> !#".
    - iSplit; iIntros "H1".
      + iDestruct ("EqO" with "H1") as (?) "[? ?]".
        iExists _. iFrame. by iApply "Hown".
      + iDestruct "H1" as (?) "[? ?]". iApply "EqO".
        iExists _. iFrame. by iApply "Hown".
    - iIntros (?). iSplit; iIntros "?".
      + iApply "Hshr". by iApply "EqS".
      + iApply "EqS". by iApply "Hshr".
  Qed.

  Lemma rwlock_proto_create l q tid α ty E (SUB: lftE ⊆ E):
    ⎡ lft_ctx ⎤ -∗
    (q).[α] -∗ ▷ ⎡ hist_inv ⎤ -∗
    &{α} ((l +ₗ 1) ↦∗{1}: ty_own ty tid)%I -∗
    &{α} (∃ n : Z, l ↦{1} #n ∗ ⌜-1 ≤ n⌝) ={E}=∗
    (q).[α] ∗ (∃ γ γs tyO tyS, rwlock_proto l γ γs α tyO tyS tid ty).
  Proof.
    iIntros "#LFT [Htok Htok'] #HINV Hvl H".
    set tyO : vProp := ((l +ₗ 1) ↦∗{1}: ty.(ty_own) tid)%I.
    set tyS : lft → vProp := (λ α', ty.(ty_shr) α' tid (l +ₗ 1))%I.
    iAssert (□ (tyO ↔ (l +ₗ 1) ↦∗{1}: ty.(ty_own) tid))%I as "#EqO".
    { iIntros "!#". by iApply (bi.iff_refl True%I). }
    iAssert (□ (∀ α', tyS α' ↔ ty.(ty_shr) α' tid (l +ₗ 1)))%I as "#EqS".
    { iIntros "!#" (?). by iApply (bi.iff_refl True%I). }
    iAssert (□ (∀ κ q' E', ⌜lftE ⊆ E'⌝ -∗ &{κ} tyO -∗
                           (q').[κ] ={E'}=∗ (□ tyS κ) ∗ (q').[κ]))%I as "#Share".
    { iIntros "!#" (????) "tyO tok".
      by iMod (ty_share with "LFT tyO tok") as "[#? $]". }
    iMod (bor_acc_cons with "LFT H Htok") as "[H Hclose]"; [done|].
    set P : gname → val → vProp :=
      (λ γs v, ∃ st, ⌜v = #(Z_of_rwlock_st st)⌝ ∗
               ⎡ own γs (st_global st : rwlockR) ⎤ ∗
               match st with
                | None => (* Not locked. *)
                      &{α}tyO
                | Some (inr (ν, q, n)) => (* Locked for read. *)
                      ∃ q', □ (1.[ν] ={↑lftN ∪ ↑histN,↑histN}▷=∗ [†ν]) ∗
                        ([†ν] ={↑lftN ∪ ↑histN}=∗ &{α}tyO) ∗
                        (□ tyS (α ⊓ ν)) ∗ ⌜(q + q')%Qp = 1%Qp⌝ ∗
                        q'.[ν]
                | _ => (* Locked for write. *) True
               end)%I.
    iAssert (|={E}=> ∃ γs, (q).[α] ∗ &{α} (∃ v, l ↦{1} v ∗ P γs v))%I
      with "[H Htok' Hvl Hclose]" as "H".
    { iDestruct "H" as ([|n|[]]) "[>Hn >%]"; try lia.
      - iMod (own_alloc (st_global None : rwlockR)) as (γs) "Hst";
          first by apply auth_auth_valid.
        iExists γs. iFrame "Htok'".
        iMod ("Hclose" with "[] [-]") as "[$ $]"; [..|done].
        + iIntros "!>". iDestruct 1 as (v) "[>Hl P]".
          iDestruct "P" as (st) "[>% _]". subst v. iExists _. iFrame "Hl".
          iIntros "!> !> !%". destruct st as [[|[[]]]|]; simpl; lia.
        + iExists _. iFrame "Hn". iExists None. by iFrame "Hst Hvl".
      - iMod (lft_create with "LFT") as (ν) "[[Htok1 Htok2] #Hhν]"; [done|].
        iMod (own_alloc (st_global (Some $ inr (ν, (1/2)%Qp, n)))) as (γs) "Hst".
        { by apply auth_auth_valid. }
        iMod (rebor _ _ (α ⊓ ν) with "LFT [] Hvl") as "[Hvl Hh]"; [done|..].
        { iApply lft_intersect_incl_l. }
        iDestruct (lft_intersect_acc with "Htok' Htok1") as (q') "[Htok Hclose2]".
        iMod (ty_share with "LFT Hvl Htok") as "[#Hshr Htok]"; [done|].
        iDestruct ("Hclose2" with "Htok") as "[$ Htok]".
        iExists γs. iMod ("Hclose" with "[] [-]") as "[$ $]"; [..|done].
        + iIntros "!>". iDestruct 1 as (v) "[>Hl P]".
          iDestruct "P" as (st) "[>% _]". subst v. iExists _. iFrame "Hl".
          iIntros "!> !> !%". destruct st as [[|[[]]]|]; simpl; lia.
        + iExists _. iFrame "Hn". iExists (Some $ inr (ν, (1/2)%Qp, n)).
          iSplitL ""; [done|]. iFrame "Hst". iExists _.
          iFrame "Hshr Htok2 Hhν". iSplitR ""; [|by rewrite Qp_div_2].
          iIntros "!> Hν". iDestruct (lft_tok_dead with "Htok Hν") as "[]".
      - iMod (own_alloc (st_global (Some $ inl ()))) as (γs) "Hst".
        { by apply auth_auth_valid. }
        iExists γs. iFrame "Htok'".
        iMod ("Hclose" with "[] [-]") as "[$ $]"; [..|done].
        + iIntros "!>". iDestruct 1 as (v) "[>Hl P]".
          iDestruct "P" as (st) "[>% _]". subst v. iExists _. iFrame "Hl".
          iIntros "!> !> !%". destruct st as [[|[[]]]|]; simpl; lia.
        + iExists _. iFrame "Hn". iExists (Some $ inl ()). iSplitL ""; [done|].
          by iFrame "Hst". }
    iMod "H" as (γs) "[Htok H]".
    iMod (GPS_aSP_Init rwlockN (rwlock_interp γs α tyO tyS) rwlock_noCAS
              l α () _ (P γs) (λ _ _ _, True)%I
              with "LFT Htok HINV H [] []") as "[$ Hproto]";
          [solve_ndisj|done|..].
    { iIntros (t γ v) "P W". iSplitR ""; [|done].
      iDestruct "P" as (st) "[>% [Hst H]]". subst v. iExists _.
      iFrame "Share Hst". iSplitL ""; [done|].
      destruct st as [[|[[] n]]|]; [done|..].
      - iDestruct (GPS_SWWriter_shared with "W") as "[$ R]".
        iIntros "!> !>". iDestruct "H" as (q') "(H† & Hinh & HtyS & #Eqq & Htok')".
        iExists q'. iFrame "H† Hinh HtyS Eqq Htok'". iDestruct "Eqq" as %Eqq.
        rewrite -Eqq. iDestruct "R" as "[? $]".
      - iFrame "H". by iDestruct (GPS_SWWriter_shared with "W") as "[$ $]". }
    { iIntros "!#" (t γ s v) "Int !> !>".
      iDestruct "Int" as (st) "(#Eq & Hshr & Hst & Int)". iExists st.
      iFrame "Eq Hst". destruct st as [[|[[] n]]|]; [done|..].
      - iDestruct "Int" as "[W Int]".
        iDestruct "Int" as (q') "(H† & Hinh & HtyS & Eqq & Htok' & R)".
        iExists q'. iFrame "H† Hinh HtyS Eqq Htok'".
      - iDestruct "Int" as "[$ _]". }
    iDestruct "Hproto" as (γ t v) "[R _]".
    iExists γ, γs, tyO, tyS. iIntros "!>".
    iFrame "EqO EqS". iExists _, _. iFrame "R".
  Qed.


  Lemma rwlock_interp_read_acq l t s v γ γs α tyO tyS tid:
    (▽{tid} rwlock_interp γs α tyO tyS false l γ t s v : vProp) -∗
      ∃ st, ⌜v = #(Z_of_rwlock_st st)⌝.
  Proof.
    iIntros "Int". iDestruct (acq_exist with "Int") as (?) "Int".
    iDestruct (acq_sep_elim with "Int") as "[Int _]".
    iDestruct (acq_pure_elim with "Int") as "?". iExists _. iFrame.
  Qed.

End rwlock_inv.

Section rwlock.
  Context `{!typeG Σ, !rwlockG Σ}.

  (* Original Rust type (we ignore poisoning):
     pub struct RwLock<T: ?Sized> {
         inner: Box<sys::RWLock>,
         poison: poison::Flag,
         data: UnsafeCell<T>,
     }
  *)

  Program Definition rwlock (ty : type) :=
    {| ty_size := S ty.(ty_size);
       ty_own tid vl :=
         (⎡ hist_inv ⎤ ∗ match vl return _ with
                         | #(LitInt z) :: vl' => ⌜-1 ≤ z⌝ ∗ ty.(ty_own) tid vl'
                         | _ => False
                         end)%I;
       ty_shr κ tid l :=
        (∃ α, κ ⊑ α ∗ ∃ γ γs tyO tyS, rwlock_proto l γ γs α tyO tyS tid ty)%I |}.
  Next Obligation.
    iIntros (??[|[[]|]]); try iIntros "[? []]". rewrite ty_size_eq.
    iIntros "[_ [_ %]] !% /=". congruence.
  Qed.
  Next Obligation.
    iIntros (ty E κ l tid q ?) "#LFT Hb Htok".
    iMod (bor_acc_cons with "LFT Hb Htok") as "[H Hclose]". done.
    iDestruct "H" as ([|[[| |n]|]vl]) "[H↦ [#hInv H]]"; try iDestruct "H" as ">[]".
    iDestruct "H" as "[>% Hown]".
    iMod ("Hclose" $! ((∃ n:Z, l ↦ #n ∗ ⌜-1 ≤ n⌝) ∗
            (l +ₗ 1) ↦∗: ty.(ty_own) tid) with "[] [-]")%I as "[H Htok]".
    { iIntros "!> [Hn Hvl] !>". iDestruct "Hn" as (n') "[Hn >%]".
      iDestruct "Hvl" as (vl') "[H↦ Hvl]".
      iExists (#n'::vl'). rewrite heap_mapsto_vec_cons. iFrame "∗% hInv". }
    { iNext. rewrite heap_mapsto_vec_cons. iDestruct "H↦" as "[Hn Hvl]".
      iSplitL "Hn"; [eauto|iExists _; iFrame]. }
    iMod (bor_sep with "LFT H") as "[Hn Hvl]". done.
    iMod (rwlock_proto_create with "LFT Htok hInv Hvl Hn") as "[$ Hproto]". done.
    iExists κ. iFrame "Hproto". by iApply lft_incl_refl.
  Qed.
  Next Obligation.
    iIntros (?????) "#Hκ H". iDestruct "H" as (α) "[#??]".
    iExists _. iFrame. iApply lft_incl_trans; auto.
  Qed.

  Global Instance rwlock_wf ty `{!TyWf ty} : TyWf (rwlock ty) :=
    { ty_lfts := ty.(ty_lfts); ty_wf_E := ty.(ty_wf_E) }.

  Global Instance rwlock_type_ne : TypeNonExpansive rwlock.
  Proof.
    constructor;
      solve_proper_core ltac:(fun _ => exact: type_dist2_S
                            || (eapply rwlock_proto_type_ne; try reflexivity)
                            || f_type_equiv || f_equiv).
  Qed.

  Global Instance rwlock_ne : NonExpansive rwlock.
  Proof.
    constructor;
      solve_proper_core ltac:(fun _ => (eapply ty_size_ne; try reflexivity)
                            || f_equiv).
  Qed.

  Global Instance rwlock_mono E L : Proper (eqtype E L ==> subtype E L) rwlock.
  Proof.
    (* TODO : this proof is essentially [solve_proper], but within the logic.
              It would easily gain from some automation. *)
    iIntros (ty1 ty2 EQ qL) "HL". generalize EQ. rewrite eqtype_unfold=>EQ'.
    iDestruct (EQ' with "HL") as "#EQ'".
    iDestruct (rwlock_proto_proper with "HL") as "#Hty1ty2"; first done.
    iIntros "!# #HE". iDestruct ("EQ'" with "HE") as "(% & #Hown & #Hshr)".
    iSplit; [|iSplit; iIntros "!# * H"].
    - iPureIntro. simpl. congruence.
    - destruct vl as [|[[]|]]; try done. iDestruct "H" as "[$ [$ H]]".
      by iApply "Hown".
    - iDestruct "H" as (α) "[Ha H]". iExists α. iFrame "Ha".
      iDestruct "H" as (????) "lc". iExists _,_,_,_.
      by iApply ("Hty1ty2" with "HE lc").
  Qed.
  Lemma rwlock_mono' E L ty1 ty2 :
    eqtype E L ty1 ty2 → subtype E L (rwlock ty1) (rwlock ty2).
  Proof. eapply rwlock_mono. Qed.
  Global Instance rwlock_proper E L : Proper (eqtype E L ==> eqtype E L) rwlock.
  Proof. by split; apply rwlock_mono. Qed.
  Lemma rwlock_proper' E L ty1 ty2 :
    eqtype E L ty1 ty2 → eqtype E L (rwlock ty1) (rwlock ty2).
  Proof. eapply rwlock_proper. Qed.

  (* Apparently, we don't need to require ty to be sync,
     contrarily to Rust's implementation. *)
  Global Instance rwlock_send :
    Send ty → Send (rwlock ty).
  Proof. move=>????[|[[]|]]//=??. iIntros "[$[$?]]". by iApply send_change_tid. Qed.

  Global Instance rwlock_sync :
    Send ty → Sync ty → Sync (rwlock ty).
  Proof.
    move=>???????/=. apply bi.exist_mono=>?. apply bi.sep_mono_r.
    do 4 apply bi.exist_mono=>?.
    apply bi.sep_mono; last apply bi.sep_mono_l;
      apply bi.later_mono; iIntros "#H !#".
    - iSplit; iIntros "H1".
      + iDestruct ("H" with "H1") as (?) "[??]".
        iExists _. iFrame. by iApply send_change_tid.
      + iApply "H". iDestruct "H1" as (?) "[??]".
        iExists _. iFrame. by iApply send_change_tid.
    - iIntros (α). iSpecialize ("H" $! α). iSplit; iIntros "H1".
      + iApply sync_change_tid. by iApply "H".
      + iApply "H". by iApply sync_change_tid.
  Qed.

End rwlock.

Hint Resolve rwlock_mono' rwlock_proper' : lrust_typing.
